import dialogic.nlg.controls as ctrl


def test_simple_button():
    data = {
        "title": 'button title'
    }
    button = ctrl.Button(**data)
    data['hide'] = True
    assert button.to_dict() == data


def test_fixed_button_with_url():
    data = {
        'title': 'button title',
        'url': 'https://example.com/',
        'hide': False
    }
    button = ctrl.Button(**data)
    assert button.to_dict() == data



"""
This package implements Yandex.Dialogs request protocol.
The official documentation is available at
https://yandex.ru/dev/dialogs/alice/doc/protocol-docpage/#request
"""

import attr
from typing import Dict, List, Optional
from dialogic.utils.serialization import FreeSerializeable, list_converter, dict_converter


class ENTITY_TYPES:
    YANDEX_GEO = 'YANDEX.GEO'
    YANDEX_FIO = 'YANDEX.FIO'
    YANDEX_NUMBER = 'YANDEX.NUMBER'
    YANDEX_DATETIME = 'YANDEX.DATETIME'


@attr.s
class Meta(FreeSerializeable):
    locale: str = attr.ib()
    timezone: str = attr.ib()
    client_id: str = attr.ib()
    interfaces: dict = attr.ib(factory=dict)

    @property
    def has_screen(self) -> bool:
        return self.interfaces and 'screen' in self.interfaces

    @property
    def has_account_linking(self) -> bool:
        return self.interfaces and 'account_linking' in self.interfaces


@attr.s
class Span(FreeSerializeable):
    start: int = attr.ib()
    end: int = attr.ib()


@attr.s
class Entity(FreeSerializeable):
    type: str = attr.ib()  # may be one of ENTITY_TYPES, but not only
    tokens: Span = attr.ib(converter=Span.from_dict)
    value = attr.ib()


@attr.s
class Slot(FreeSerializeable):
    type: str = attr.ib()
    tokens: Optional[Span] = attr.ib(converter=Span.from_dict, default=None)
    value = attr.ib(default=None)


@attr.s
class Intent(FreeSerializeable):
    slots: Dict[str, Slot] = attr.ib(converter=dict_converter(Slot))

    def get_form(self) -> Dict[str, str]:
        return {
            slot_name: slot.value
            for slot_name, slot in self.slots.items()
        }


@attr.s
class NLU(FreeSerializeable):
    tokens: List[str] = attr.ib(factory=list)
    entities: List[Entity] = attr.ib(converter=list_converter(Entity), factory=list)
    intents: Dict[str, Intent] = attr.ib(converter=dict_converter(Intent), factory=dict)

    def get_forms(self) -> Dict[str, Dict[str, str]]:
        return {
            intent_name: intent.get_form()
            for intent_name, intent in self.intents.items()
        }


@attr.s
class Request(FreeSerializeable):
    command: str = attr.ib(default=None)
    original_utterance: str = attr.ib(default=None)
    type: str = attr.ib(default=None)
    markup = attr.ib(factory=dict)
    payload = attr.ib(factory=dict)
    nlu: Optional[NLU] = attr.ib(converter=NLU.from_dict, default=None)
    show_type: Optional[str] = attr.ib(default=None)


class User(FreeSerializeable):
    user_id: str = attr.ib(default=None)
    access_token: str = attr.ib(default=None)


class Application(FreeSerializeable):
    application_id: str = attr.ib(default=None)


@attr.s
class Location(FreeSerializeable):
    lat: Optional[float] = attr.ib(default=None)
    lon: Optional[float] = attr.ib(default=None)
    accuracy: Optional[float] = attr.ib(default=None)


@attr.s
class Session(FreeSerializeable):
    message_id: int = attr.ib()
    session_id: str = attr.ib()
    skill_id: str = attr.ib()
    # user_id is deprecated, use application.application_id or user.user_id instead
    user_id: str = attr.ib()
    user: Optional[User] = attr.ib(default=None)
    application: Optional[Application] = attr.ib(default=None)
    new: bool = attr.ib(default=False)
    location: Optional[Location] = attr.ib(default=None, converter=Location.from_dict)


@attr.s
class State(FreeSerializeable):
    session: Optional[Dict] = attr.ib(default=None)
    user: Optional[Dict] = attr.ib(default=None)
    application: Optional[Dict] = attr.ib(default=None)


@attr.s
class YandexRequest(FreeSerializeable):
    meta: Meta = attr.ib(converter=Meta.from_dict)
    request: Request = attr.ib(converter=Request.from_dict)
    session: Optional[Session] = attr.ib(converter=Session.from_dict)
    new: bool = attr.ib(default=False)
    version: str = attr.ib(default='1.0')
    state: Optional[State] = attr.ib(converter=State.from_dict, default=None)




def ensure_mongo_index(index_name, collection, unique=False, index_type='hashed'):
    if index_name not in collection.index_information():
        collection.create_index([(index_name, index_type)], unique=unique)



import re
from typing import Dict

try:
    import regex
except ImportError:
    regex = None

regex_or_re = regex or re


def drop_none(d):
    return {k: v for k, v in d.items() if v is not None}


def match_forms(text: str, intents: dict) -> Dict[str, Dict]:
    forms = {}
    for intent_name, intent_value in intents.items():
        if 'regexp' in intent_value:
            expressions = intent_value['regexp']
            if isinstance(expressions, str):
                expressions = [expressions]
            for exp in expressions:
                match = regex_or_re.match(exp, text)
                if match:
                    forms[intent_name] = drop_none(match.groupdict())
                    break
    return forms



import copy
import logging
from typing import Optional, Dict

from ..dialog.serialized_message import SerializedMessage
from ..dialog import Context, Response


logger = logging.getLogger(__name__)


class BaseAdapter:
    """ A base class for adapters that encode and decode messages into a single format """

    def make_context(self, message, **kwargs) -> Context:
        """ Get a raw `message` in a platform-specific format and encode it into a unified `Context` """
        raise NotImplementedError()

    def make_response(self, response: Response, original_message=None, **kwargs):
        """ Get a unified `Response` object and decode it into the platform-specific response """
        raise NotImplementedError()

    def serialize_context(self, context: Context, data=None, **kwargs) -> Optional[SerializedMessage]:
        if data is None:
            data = context.raw_message
        if context.request_id is not None:
            kwargs['request_id'] = context.request_id
        return SerializedMessage(
            text=context.message_text,
            user_id=context.user_id,
            from_user=True,
            data=data,
            source=context.source,
            session_id=context.session_id,
            **kwargs
        )

    def serialize_response(self, data, context: Context, response: Response, **kwargs) -> Optional[SerializedMessage]:
        data = copy.deepcopy(data)
        if context.request_id is not None:
            kwargs['request_id'] = context.request_id
        if response.label:
            kwargs['label'] = response.label
        return SerializedMessage(
            text=response.text,
            user_id=context.user_id,
            from_user=False,
            data=data,
            source=context.source,
            session_id=context.session_id,
            **kwargs
        )

    def uses_native_state(self, context: Context) -> bool:
        """ Whether dialog state can be extracted directly from a context"""
        return False

    def get_native_state(self, context: Context) -> Optional[Dict]:
        """ Return native dialog state if it is possible"""
        return



import re

from .base import CascadableDialogManager, Context, Response
from dialogic.nlu import basic_nlu, matchers
from dialogic.utils.configuration import load_config


class FieldConfig:
    def __init__(self, obj, idx=-1):
        self.validate_regexp = re.compile(obj.get('validate_regexp', '.*'))
        self.id = id
        self.name = obj.get('name', 'field_{}'.format(idx))
        self.question = obj['question']
        self.validate_message = obj.get('validate_message')
        self.options = obj.get('options')
        self.suggests = obj.get('suggests')

        if self.options is not None:
            self.matcher = matchers.make_matcher(**obj.get('matching', {'key': 'levenshtein', 'threshold': 0.8}))
            self.matcher.fit(self.options, self.options)
        else:
            self.matcher = None


class FormConfig:
    def __init__(self, config):
        self._cfg = load_config(config)
        # todo: validate everything
        self.form_name = self._cfg['form_name']

        self.start_regex = re.compile(self._cfg['start'].get('regexp', '.*'))
        self.start_message = self._cfg['start'].get('message')
        self.start_suggests = self._cfg['start'].get('suggests', [])

        exit_block = self._cfg.get('exit', {})
        self.exit_regexp = re.compile(exit_block['regexp']) if 'regexp' in exit_block else None
        self.exit_message = exit_block.get('message')
        self.exit_suggest = exit_block.get('suggest')

        self.finish_message = self._cfg.get('finish', {}).get('message')

        self.fields = [FieldConfig(obj, idx=i) for i, obj in enumerate(self._cfg['fields'])]
        self.default_field = FieldConfig(self._cfg['default_field']) if 'default_field' in self._cfg else None
        self.num_fields = len(self.fields)


class FormFillingDialogManager(CascadableDialogManager):
    def __init__(self, config, *args, **kwargs):
        super(FormFillingDialogManager, self).__init__(*args, **kwargs)
        self.config = FormConfig(config)

    def try_to_respond(self, ctx: Context):
        user_object = ctx.user_object or {}
        normalized = basic_nlu.fast_normalize(ctx.message_text)
        form = user_object.get('forms', {}).get(self.config.form_name, {})
        if form.get('is_active'):
            if self.config.exit_regexp and re.match(self.config.exit_regexp, normalized):
                form['is_active'] = False
                return Response(text=self.config.exit_message, user_object=user_object)
            question_id = form['next_question']
            validated_answer = self.validate_answer(form, ctx.message_text)
            if validated_answer is not None:
                form['fields'][self.config.fields[question_id].name] = validated_answer
                next_question_id = question_id + 1
                if next_question_id >= self.config.num_fields:
                    form.pop('next_question')
                    form['is_active'] = False
                    result = self.handle_completed_form(form, user_object, ctx)
                    if result is not None:
                        return result
                    return Response(text=self.config.finish_message, user_object=user_object)
                form['next_question'] = next_question_id
                return self.ask_question(next_question_id, user_object=user_object, reask=False)
            else:
                return self.ask_question(question_id, user_object=user_object, reask=True)
        elif re.match(self.config.start_regex, normalized):
            return self.start_dialogue(ctx)
        return None

    def start_dialogue(self, ctx: Context):
        """ Initialize the form, and say the intro or ask the first question.
        This function is taken out of try_to_respond, so that it could be triggered by some external factor.
        """
        user_object = ctx.user_object or {}
        if 'forms' not in user_object:
            user_object['forms'] = {}
        user_object['forms'][self.config.form_name] = {
            'fields': {},
            'is_active': True,
            'next_question': -1
        }
        form = user_object['forms'][self.config.form_name]
        if self.config.start_message is not None:
            form['next_question'] = -1
            return Response(
                text=self.config.start_message, suggests=self.config.start_suggests, user_object=user_object
            )
        else:
            form['next_question'] = 0
            return self.ask_question(0, user_object, reask=False)

    def ask_question(self, next_question_id, user_object, reask=False):
        the_question = self.config.fields[next_question_id]
        response = the_question.question
        if reask:
            if the_question.validate_message is not None:
                response = the_question.validate_message
            elif 'validate_message' in self.config.default_field:
                response = self.config.default_field['validate_message']
        if the_question.options is not None:
            suggests = the_question.options
        elif the_question.suggests is not None:
            suggests = the_question.suggests
        else:
            suggests = []
        if self.config.exit_suggest is not None:
            suggests.append(self.config.exit_suggest)
        return Response(user_object=user_object, text=response, suggests=suggests)

    def validate_answer(self, form, message_text):
        question_id = form.get('next_question', -1)
        if question_id == -1:
            return message_text
        the_question = self.config.fields[question_id]
        normalized_text = basic_nlu.fast_normalize(message_text)
        if the_question.options is not None:
            winner_label, best_score = the_question.matcher.match(message_text)
            return winner_label
        if the_question.validate_regexp is not None:
            if re.match(the_question.validate_regexp, normalized_text):
                return message_text
            else:
                return None
        return message_text

    def handle_completed_form(self, form, user_object, ctx):
        """ This method can be overwritten to do something useful.
        If it is not overwritten, then the dialog ends by the `finish_message` from config.
        If it is overwritten, then it should probably put the updated `user_object` argument into its Response.
        """
        return None



from .base import BaseAdapter
from .alice import AliceAdapter
from .facebook import FacebookAdapter
from .text import TextAdapter
from .tg import TelegramAdapter
from .vk import VkAdapter
from .salut import SalutAdapter



from typing import Dict

from dialogic.dialog.names import SOURCES, REQUEST_TYPES
from dialogic.adapters.base import BaseAdapter, Context, Response, logger


class TextAdapter(BaseAdapter):
    SOURCE = SOURCES.TEXT

    def make_context(self, message: Dict, **kwargs) -> Context:
        ctx = Context(
            user_object=None,
            message_text=message,
            metadata={},
            user_id='the_text_user',
            session_id='the_text_session',
            source=self.SOURCE,
            raw_message=message,
        )
        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        result = response.text
        if response.voice is not None and response.voice != response.text:
            result = result + '\n[voice: {}]'.format(response.voice)
        if response.image_id:
            result = result + '\n[image: {}]'.format(response.image_id)
        if response.image_url:
            result = result + '\n[image: {}]'.format(response.image_url)
        if response.sound_url:
            result = result + '\n[sound: {}]'.format(response.sound_url)
        if len(response.links) > 0:
            result = result + '\n' + ', '.join(
                ['[{}: {}]'.format(link['title'], link['url']) for link in response.links]
            )
        if len(response.suggests) > 0:
            result = result + '\n' + ', '.join(['[{}]'.format(s) for s in response.suggests])
        if len(response.commands) > 0:
            result = result + '\n' + ', '.join(['{{{}}}'.format(c) for c in response.commands])
        return [result, response.has_exit_command]



import copy
import logging
import uuid

from typing import Optional

from .names import REQUEST_TYPES, SOURCES
from dialogic.interfaces.yandex import YandexRequest


logger = logging.getLogger(__name__)


class Context:
    def __init__(
            self, user_object, message_text, metadata, request_id=None, user_id=None, source=None, raw_message=None,
            request_type=REQUEST_TYPES.SIMPLE_UTTERANCE, payload=None, yandex=None,
            session_id=None,
    ):
        self._user_object = copy.deepcopy(user_object)
        self.message_text = message_text
        self.metadata = metadata
        self.request_id = request_id or str(uuid.uuid1())
        self.user_id = user_id
        self.session_id = session_id
        self.source = source
        self.raw_message = raw_message
        self.request_type = request_type
        self.payload = payload
        self.yandex: Optional[YandexRequest] = yandex

    @property
    def user_object(self):
        # todo: make this object explicitly frozen
        return self._user_object

    def add_user_object(self, user_object):
        self._user_object = copy.deepcopy(user_object)

    def session_is_new(self):
        # todo: define new session for non-Alice sources as well
        return bool(self.metadata.get('new_session'))

    @classmethod
    def from_raw(cls, source, message):
        raise DeprecationWarning('This method is not used anymore. Please use the adapters subpackage instead.')



from unittest.mock import patch

import dialogic

from dialogic.utils.content_manager import YandexImageAPI
from dialogic.dialog_connector import DialogConnector


def test_image_api():
    token = '1234'
    skill_id = '5678'
    default_image_id = '9101112'
    manager = YandexImageAPI(token=token, skill_id=skill_id, default_image_id=default_image_id)

    # test updating
    with patch('requests.get') as mock:
        mock.return_value.json = lambda: {
            'images': [
                {'id': 'tree', 'origUrl': 'trees.com/tree.jpg', 'size': 100500, 'createdAt': 'today'},
                {'id': 'snake', 'size': 100500, 'createdAt': 'yesterday'},
            ],
            'total': 2
        }
        manager.update_images()
        assert mock.called
        args = mock.call_args[1]
        assert args['url'] == 'https://dialogs.yandex.net/api/v1/skills/{}/images'.format(skill_id)
        assert args['headers'] == {'Authorization': 'OAuth 1234'}

        assert len(manager.url2image) == 1
        assert len(manager.id2image) == 2

    # test usage without upload
    with patch('requests.post') as mock:
        assert manager.get_image_id_by_url('trees.com/tree.jpg') == 'tree'
        assert manager.get_image_id_by_url('trees.com/pine.jpg') == default_image_id
        assert not mock.called

    # test adding an existing image
    with patch('requests.post') as mock:
        image = manager.add_image(url='trees.com/tree.jpg')
        assert not mock.called
        assert image.id == 'tree'

    # test uploading an image
    with patch('requests.post') as mock:
        oak_url = 'trees.com/oak.jpg'
        mock.return_value.json = lambda: {
            'image': {
                'id': 'oak',
                'origUrl': oak_url,
                'size': 100500,
                'createdAt': 'right_now'
            }
        }
        image = manager.add_image(url=oak_url)
        assert mock.called
        assert mock.call_args[1]['url'] == 'https://dialogs.yandex.net/api/v1/skills/{}/images'.format(skill_id)
        assert image.origUrl == oak_url
        assert image.id == 'oak'

        assert len(manager.id2image) == 3
        assert len(manager.url2image) == 2

    # test uploading an image on the fly
    manager.upload_just_in_time = True
    with patch('requests.post') as mock:
        maple_url = 'trees.com/maple.jpg'
        mock.return_value.ok = True
        mock.return_value.json = lambda: {
            'image': {
                'id': 'maple',
                'origUrl': maple_url,
                'size': 100500,
                'createdAt': 'right_now'
            }
        }
        assert manager.get_image_id_by_url(maple_url) == 'maple'
        assert mock.called

        assert len(manager.id2image) == 4
        assert len(manager.url2image) == 3

    # test removing an image
    with patch('requests.delete') as mock:
        mock.return_value.ok = True
        mock.return_value.json = lambda: {'result': 'ok'}

        manager.delete_image('maple')
        assert mock.called
        assert mock.call_args[1]['url'] == 'https://dialogs.yandex.net/api/v1/skills/{}/images/maple'.format(skill_id)

        assert len(manager.id2image) == 3
        assert len(manager.url2image) == 2

    # test automatic conversion of urls to ids
    connector = DialogConnector(dialog_manager=None, image_manager=manager)
    response = dialogic.dialog.Response('this is an oak', image_url='trees.com/oak.jpg')
    result = connector.standardize_output(
        source=dialogic.SOURCES.ALICE,
        original_message={'version': '1.0'},
        response=response
    )
    assert result['response']['card'] == {
        'type': 'BigImage',
        'image_id': 'oak',
        'description': 'this is an oak',
    }



from dialogic.nlg.morph import inflect_case, with_number, human_duration


def test_inflect_case():
    assert inflect_case('жук', 'datv') == 'жуку'


def test_with_number():
    assert with_number('слон', 0) == '0 слонов'
    assert with_number('слон', 1) == '1 слон'
    assert with_number('слон', 2) == '2 слона'
    assert with_number('слон', 5) == '5 слонов'


def test_duration():
    assert human_duration(hours=1, minutes=2, seconds=5) == '1 час 2 минуты 5 секунд'
    assert human_duration() == '0 секунд'
    assert human_duration(minutes=22) == '22 минуты'



from typing import Dict

from dialogic.dialog.names import SOURCES, REQUEST_TYPES
from dialogic.adapters.base import BaseAdapter, Context, Response, logger


class FacebookAdapter(BaseAdapter):
    SOURCE = SOURCES.FACEBOOK

    def make_context(self, message: Dict, **kwargs) -> Context:
        uid = self.SOURCE + '__' + message['sender']['id']
        ctx = Context(
            user_object=None,
            message_text=message.get('message', {}).get('text', ''),
            metadata={},
            user_id=uid,
            session_id=uid,
            source=self.SOURCE,
            raw_message=message,
        )

        if not message.get('message', {}).get('text', ''):
            payload = message.get('postback', {}).get('payload')
            if payload is not None:
                ctx.payload = payload
                ctx.request_type = REQUEST_TYPES.BUTTON_PRESSED  # todo: check if it is really the case
            # todo: do something in case of attachments (message['message'].get('attachments'))
            # if user sends us a GIF, photo,video, or any other non-text item

        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        if response.raw_response is not None:
            return response.raw_response
        result = {'text': response.text}
        if response.suggests or response.links:
            links = [{'type': 'web_url', 'title': link['title'], 'url': link['url']} for link in response.links]
            suggests = [{'type': 'postback', 'title': s, 'payload': s} for s in response.suggests]
            result = {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": response.text,
                        "buttons": links + suggests
                    }
                }
            }
        return result  # for bot.send_message(recipient_id, result)



# -*- coding: utf-8 -*-
from dialogic import adapters, cascade, dialog, dialog_manager, interfaces, \
    nlg, nlu, storage, testing, utils, criteria, dialog_connector
from dialogic.server import flask_server
from dialogic.storage import session_storage, message_logging
from dialogic.dialog_manager.base import COMMANDS
from dialogic.nlu import basic_nlu

from dialogic.dialog.names import COMMANDS, REQUEST_TYPES, SOURCES



import json
import logging
import random
import re
import requests
import threading
import time


logger = logging.getLogger(__name__)

VK_API_URL = 'https://api.vk.com/method/'


class VKBot:
    def __init__(self, token, group_id, api_version='5.103', polling_freq=5, polling_wait=25, webhook_key=None):
        self.token = token
        self.group_id = group_id
        self.api_version = api_version
        self.message_handlers = []
        self._polling_key = None
        self._polling_server = None
        self._polling_ts = None
        self.polling_freq = polling_freq
        self.polling_wait = polling_wait
        self.webhook_key = webhook_key

    def polling(self):
        """ Start long polling with wait timeout self.polling_wait and interval self.polling_freq """
        self.set_polling_server()
        while True:
            updates = self.retrieve_updates()
            logging.debug('got {} new updates from polling'.format(len(updates)))
            self.process_new_updates(updates=updates)
            time.sleep(self.polling_freq)

    def set_polling_server(self):
        """ Prepare the requisites for long polling """
        j = self._request_api('groups.getLongPollServer')['response']
        self._polling_server = j['server']
        self._polling_key = j['key']
        self._polling_ts = j['ts']

    def retrieve_updates(self):
        """ Do one iteration of long polling and return a list of fresh updates """
        assert self._polling_server is not None
        logger.info('Start retrieving polling updates...')
        result = requests.get(
            url=self._polling_server,
            params={
                'act': 'a_check',
                'key': self._polling_key,
                'ts': self._polling_ts,
                'wait': self.polling_wait,
                'v': self.api_version,  # it's undocumented, so I'm not sure it always works as expected
            },
        )
        self._fail_on_bad_response('retrieve long poll', result)
        j = result.json()
        self._polling_ts = j['ts']
        updates = j['updates']  # each update looks like {'type': update_type, 'object': the message}
        return updates

    def process_new_updates(self, updates):
        """ Take a list of raw updates, and apply the first relevant handler to each one """
        for update in updates:
            logger.debug('processing an update: {}'.format(update))
            for handler in self.message_handlers:
                if self._apply_filters(filters=handler['filters'], update=update):
                    handler['function'](update)
                    break  # only one handler is supposed to be triggered

    @staticmethod
    def _apply_filters(filters, update):
        for filter_type, filter_value in filters.items():
            if filter_value is None:
                continue
            if filter_type == 'types':
                if update.get('type') not in filter_value:
                    return False
            elif filter_type == 'regexp':
                if not re.search(filter_value, extract_message_text(update), re.IGNORECASE):
                    return False
            elif filter_type == 'func':
                if not filter_value(update):
                    return False
            else:
                raise ValueError('Filter type {} is invalid'.format(filter_type))
        return True

    def remove_webhook(self, webhook_id=None):
        """ Remove all callback webhooks from the group (default), or one or several specific webhooks """
        if webhook_id is None:
            result = self._request_api('groups.getCallbackServers')
            if 'response' in result:
                webhook_ids = [server['id'] for server in result['response']['items']]
            else:
                logger.debug(f'vk callback servers response is {result}')
                webhook_ids = []
        elif isinstance(webhook_id, int):
            webhook_ids = [webhook_id]
        else:
            # assume it's a list of ids
            webhook_ids = webhook_id

        for server_id in webhook_ids:
            logger.debug('removing server {}'.format(server_id))
            self._request_api('groups.deleteCallbackServer', server_id=server_id)

    def _get_webhook_secret_code(self):
        logger.info('looking for webhook secret code...')
        j = self._request_api('groups.getCallbackConfirmationCode')
        if 'response' in j:
            return j['response']['code']
        else:
            logger.debug(f'vk webhook secret code response is {j}')
            return

    def set_webhook(self, url, secret_key=None, title='default hook', remove_old=True, events=None):
        """ Create a webhook to the specified URL.
        By default, it is triggered by new messages, but a map of triggers may be set in `events`.
        By default, all the existing webhooks are removed, but it can be turned off by setting `remove_old=False`.
        The list of triggers is given at https://vk.com/dev/groups.setCallbackSettings
        """
        if remove_old:
            # todo: deal with concurrency (e.g. werkzeug restarts)
            self.remove_webhook()
        if secret_key is None:
            secret_key = self.webhook_key
        if secret_key is None:
            secret_key = self._get_webhook_secret_code()
            self.webhook_key = secret_key
        if events is None:
            events = {'message_new': 1}

        logger.info('setting webhook...')
        assert len(title) < 14
        j = self._request_api('groups.addCallbackServer', url=url, title=title, secret_key=secret_key)
        logger.info('webhook creation data: {}'.format(j))
        server_id = j['response']['server_id']

        logger.info('attaching incoming messages to webhook...')
        return self._request_api(
            'groups.setCallbackSettings',
            server_id=server_id,
            api_version=self.api_version,  # version of the proposed callback api
            secret_key=secret_key,
            **events
        )

    def set_postponed_webhook(self, url, secret_key=None, interval=1, remove_old=True):
        """ Run `set_webhook`, as soon as the webhook url starts responding """
        def runner():
            while True:
                logger.info('checking webhook availability...')
                result = requests.post(url, json={'type': 'confirmation', 'group_id': self.group_id, 'fake': True})
                if result.status_code == 200:
                    break
                time.sleep(interval)
            self.set_webhook(url=url, secret_key=secret_key, remove_old=remove_old)
        thread = threading.Thread(target=runner, daemon=True)
        thread.start()

    def message_handler(self, types=None, regexp=None, func=None):
        """ This decorator registers the function as a handler for certain type of messages """
        if types is None:
            types = ['message_new']

        def decorator(handler):
            def new_handler(raw_message):
                if isinstance(raw_message, VKMessage):
                    new_message = raw_message
                else:
                    new_message = VKMessage.from_json(raw_message)
                return handler(new_message)

            self.message_handlers.append({
                'function': new_handler,
                'filters': {'types': types, 'regexp': regexp, 'func': func}
            })
            return new_handler
        return decorator

    def send_message(self, user_id, text, keyboard=None, reply_to=None):
        """ Send the message with the specified text to the specified user. """
        extras = {}
        if keyboard is not None:
            if not isinstance(keyboard, str):
                keyboard = json.dumps(keyboard)
            extras['keyboard'] = keyboard
        if reply_to is not None:
            extras['reply_to'] = reply_to
        result = self._request_api(
            'messages.send',
            request_method='POST',
            user_id=user_id,
            message=text,
            random_id=random.randint(0, 1_000_000_000_000_000_000),
            group_id=None,  # pass it explicitly, to avoid confusion with user_id
            **extras,
        )
        logger.debug('message sending result: {}'.format(result))
        return result

    def _request_api(self, method, request_method='GET', **params):
        special_params = {
            'access_token': self.token,
            'v': self.api_version,
            'group_id': self.group_id,
        }
        for k, v in special_params.items():
            if k not in params:
                params[k] = v
        payload = {k: v for k, v in params.items() if v is not None}
        url = VK_API_URL + method
        if request_method == 'GET':
            result = requests.get(url=url, params=payload)
        else:  # assume it is 'POST'
            result = requests.post(url=url, data=payload)
        self._fail_on_bad_response(method, result)
        return result.json()

    @staticmethod
    def _fail_on_bad_response(request_goal, response):
        if response.status_code != 200:
            raise ValueError('Request to {} returned error code {}, with contents "{}".'.format(
                request_goal, response.status_code, response.text
            ))

    def process_webhook_data(self, data):
        """ Take the object passed into webhook, process updates, and return the appropriate answer and code
        The webhook answer serves two roles:
            1. When `type=confirmation`, it confirms its validity to the VK backend when registering webhook
            2. Otherwise, it pushes new updates into the message handlers (or elsewhere)
        """
        logger.debug('Got callback data: {}'.format(data))
        if data.get('type') == 'confirmation' and str(data.get('group_id')) == str(self.group_id):
            logger.debug('Telling the callback secret to VK!')
            return self.webhook_key or '', 200
        self.process_new_updates([data])
        return 'ok', 200  # todo: async?

    def register_flask_webhook(self, app, endpoint):
        """ Attach its own webhook to the given endpoint of the given flask app """
        from flask import request
        app.route(endpoint, methods=['POST'])(lambda: self.process_webhook_data(request.json))


def extract_message_text(update):
    # todo: extract message text in a consistent way
    message = update['object']
    if 'message' in message:
        message = message['message']
    return message.get('body') or message.get('text') or ''


class VKMessage:
    """
    Source: personal messages, like https://vk.com/dev/objects/message -
    When it comes from webhook, it looks like
        {"type": "message_new", "object": object, "group_id": 12345}
    where object is
        {"message": personal_message, "client_info": client_info}
    when it comes from long poll, it looks like the old version of API, so the format is different
    # todo: add message_id, attachments, date, etc.
    """
    def __init__(self, user_id, text):
        self.user_id = user_id
        self.text = text

    @classmethod
    def from_json(cls, data):
        message = data['object']

        if 'message' in message:
            message = message['message']

        text = message.get('body') or message.get('text') or ''
        user_id = message.get('user_id') or message.get('from_id')
        return cls(user_id=user_id, text=text)

    def to_json(self):
        return {'text': self.text, 'user_id': self.user_id}



from dialogic.nlu.matchers import make_matcher_with_regex, JaccardMatcher
from dialogic.nlu.regex_expander import load_intents_with_replacement


def test_expander():
    intents = load_intents_with_replacement(
        'tests/test_nlu/intents.yaml',
        'tests/test_nlu/expressions.yaml'
    )
    matcher = make_matcher_with_regex(JaccardMatcher(), intents=intents)
    assert matcher.match('take number 10')



import copy

from typing import Optional, Dict

from .storage.message_logging import BaseMessageLogger

from .adapters import (
    AliceAdapter, BaseAdapter, FacebookAdapter, TextAdapter, TelegramAdapter, VkAdapter, SalutAdapter
)

from dialogic.storage.session_storage import BaseStorage
from dialogic.dialog_manager.base import Response, Context
from dialogic.dialog.names import SOURCES
from dialogic.utils.content_manager import YandexImageAPI


class DialogConnector:
    """ This class provides unified interface for Telegram and Alice, and other applications """
    def __init__(
            self,
            dialog_manager,
            storage=None,
            log_storage: Optional[BaseMessageLogger] = None,
            default_source=SOURCES.TELEGRAM,
            tg_suggests_cols=1,
            alice_native_state=False,
            image_manager: Optional[YandexImageAPI] = None,
            adapters: Optional[Dict[str, BaseAdapter]] = None,
    ):
        """
        paramaters:
        - alice_native_state: bool or 'user' or 'state'
        """
        self.dialog_manager = dialog_manager
        self.default_source = default_source
        self.storage = storage or BaseStorage()
        self.log_storage: Optional[BaseMessageLogger] = log_storage  # noqa
        self.tg_suggests_cols = tg_suggests_cols
        self.alice_native_state = alice_native_state
        self.image_manager = image_manager
        self.adapters: Dict[str, BaseAdapter] = adapters or {}
        self._add_default_adapters()

    def _add_default_adapters(self):
        if SOURCES.ALICE not in self.adapters:
            self.adapters[SOURCES.ALICE] = AliceAdapter(
                native_state=self.alice_native_state,
                image_manager=self.image_manager,
            )
        if SOURCES.FACEBOOK not in self.adapters:
            self.adapters[SOURCES.FACEBOOK] = FacebookAdapter()
        if SOURCES.TEXT not in self.adapters:
            self.adapters[SOURCES.TEXT] = TextAdapter()
        if SOURCES.TELEGRAM not in self.adapters:
            self.adapters[SOURCES.TELEGRAM] = TelegramAdapter(suggest_cols=self.tg_suggests_cols)
        if SOURCES.VK not in self.adapters:
            self.adapters[SOURCES.VK] = VkAdapter()
        if SOURCES.SALUT not in self.adapters:
            self.adapters[SOURCES.SALUT] = SalutAdapter()

    def add_adapter(self, name: str, adapter: BaseAdapter):
        self.adapters[name] = adapter

    def respond(self, message, source=None):
        adapter = self.adapters.get(source)
        # todo: support different triggers - not only messages, but calendar events as well
        context = self.make_context(message=message, source=source)
        old_user_object = copy.deepcopy(context.user_object)
        if adapter and self.log_storage is not None:
            logged = adapter.serialize_context(context=context)
            if logged:
                self.log_storage.log_data(data=logged, context=context)

        response = self.dialog_manager.respond(context)
        if response.updated_user_object is not None and response.updated_user_object != old_user_object:
            if adapter and adapter.uses_native_state(context=context):
                pass  # user object is added right to the response
            else:
                self.set_user_object(context.user_id, response.updated_user_object)

        result = self.standardize_output(source=source, original_message=context.raw_message, response=response)
        if self.log_storage is not None:
            logged = self.adapters[source].serialize_response(data=result, context=context, response=response)
            if logged:
                self.log_storage.log_data(data=logged, context=context, response=response)
        return result

    def make_context(self, message, source=None):
        if source is None:
            source = self.default_source
        assert source in self.adapters, f'Source "{source}" is not in the list of dialog adapters ' \
                                        f'{list(self.adapters.keys())}.'
        adapter = self.adapters[source]
        context = adapter.make_context(message=message)

        if adapter.uses_native_state(context=context):
            user_object = adapter.get_native_state(context=context)
        else:
            user_object = self.get_user_object(context.user_id)
        context.add_user_object(user_object)
        return context

    def get_user_object(self, user_id):
        if self.storage is None:
            return {}
        return self.storage.get(user_id)

    def set_user_object(self, user_id, user_object):
        if self.storage is None:
            raise NotImplementedError()
        self.storage.set(user_id, user_object)

    def standardize_output(self, source, original_message, response: Response):

        assert source in self.adapters, f'Source "{source}" is not in the list of dialog adapters ' \
                                        f'{list(self.adapters.keys())}.'
        return self.adapters[source].make_response(response=response, original_message=original_message)

    def serverless_alice_handler(self, alice_request, context):
        """ This method can be set as a hanlder if the skill is deployed as a Yandex.Cloud Serverless Function """
        return self.respond(alice_request, source=SOURCES.ALICE)



from dialogic.nlg.sampling import sample


def test_sample():
    pattern = '{cat|dog}{ never| always}* says {bow-wow|meow}'
    variants = {sample(pattern, rnd=True) for i in range(1000)}
    assert 'cat never says bow-wow' in variants
    assert 'dog says meow' in variants
    assert len(variants) == 2 * 3 * 2
    assert len({sample(pattern, rnd=False) for i in range(1000)}) == 1



import pytest
import dialogic
import telebot

from dialogic.storage.database_utils import get_mongo_or_mock


@pytest.fixture
def empty_db():
    database = get_mongo_or_mock()
    database.get_collection('message_logs').drop()
    return database


def test_text_logging_with_connector(empty_db):
    database = empty_db
    input_message = 'hello bot'
    expected_response = 'This is the default message'
    dm = dialogic.dialog_manager.BaseDialogManager(default_message=expected_response)
    connector = dialogic.dialog_connector.DialogConnector(
        dialog_manager=dm,
        log_storage=dialogic.storage.message_logging.MongoMessageLogger(database=database)
    )
    text_response = connector.respond(message=input_message, source=dialogic.SOURCES.TEXT)

    collection = database.get_collection('message_logs')
    logs = list(collection.find())
    assert len(logs) == 2
    first, second = logs

    assert first['text'] == input_message
    assert second['text'] == expected_response

    assert first['source'] == dialogic.SOURCES.TEXT
    assert second['source'] == dialogic.SOURCES.TEXT

    assert first['from_user'] is True
    assert second['from_user'] is False

    assert first['data'] == input_message
    assert second['data'] == text_response


def test_alice_logging_with_connector(empty_db):
    input_message = {
        "meta": {
            "locale": "ru-RU",
            "timezone": "Europe/Moscow",
            "client_id": "ru.yandex.searchplugin/5.80 (Samsung Galaxy; Android 4.4)",
            "interfaces": {
                "screen": {}
            }
        },
        "request": {
            "command": "привет",
            "original_utterance": "привет",
            "type": "SimpleUtterance",
            "markup": {
                "dangerous_context": False
            },
            "payload": {},
            "nlu": {
                "tokens": ["привет"],
                "entities": []
            }
        },
        "session": {
            "new": False,
            "message_id": 4,
            "session_id": "2eac4854-fce721f3-b845abba-20d60",
            "skill_id": "3ad36498-f5rd-4079-a14b-788652932056",
            "user_id": "AC9WC3DF6FCE052E45A4566A48E6B7193774B84814CE49A922E163B8B29881DC"
        },
        "version": "1.0"
    }
    database = empty_db
    expected_response_text = 'This is the default message'
    input_message_text = input_message['request']['command']
    dm = dialogic.dialog_manager.BaseDialogManager(default_message=expected_response_text)
    connector = dialogic.dialog_connector.DialogConnector(
        dialog_manager=dm,
        log_storage=dialogic.storage.message_logging.MongoMessageLogger(database=database)
    )
    alice_response = connector.respond(message=input_message, source=dialogic.SOURCES.ALICE)

    collection = database.get_collection('message_logs')
    logs = list(collection.find())
    assert len(logs) == 2
    first, second = logs

    assert first['text'] == input_message_text
    assert second['text'] == expected_response_text

    assert first['source'] == dialogic.SOURCES.ALICE
    assert second['source'] == dialogic.SOURCES.ALICE

    assert first['from_user'] is True
    assert second['from_user'] is False

    assert first['data'] == input_message
    assert second['data'] == alice_response

    assert first['request_id'] == second['request_id']
    assert first['request_id'] is not None


def test_tg_logging_with_connector(empty_db):
    input_message_text = 'привет'
    input_message = telebot.types.Message(
        message_id=123,
        from_user=telebot.types.User(id=456, first_name='Bob', is_bot=False),
        chat=telebot.types.Chat(username='Bobby', id=123, type='private'),
        date=None, content_type='text', json_string=None,
        options={'text': input_message_text}
    )
    database = empty_db
    expected_response_text = 'This is the default message'
    dm = dialogic.dialog_manager.BaseDialogManager(default_message=expected_response_text)
    connector = dialogic.dialog_connector.DialogConnector(
        dialog_manager=dm,
        log_storage=dialogic.storage.message_logging.MongoMessageLogger(database=database)
    )
    tg_response = connector.respond(message=input_message, source=dialogic.SOURCES.TELEGRAM)

    if 'reply_markup' in tg_response:
        tg_response['reply_markup'] = tg_response['reply_markup'].to_json()

    collection = database.get_collection('message_logs')
    logs = list(collection.find())
    assert len(logs) == 2
    first, second = logs

    assert first['text'] == input_message_text
    assert second['text'] == expected_response_text

    assert first['source'] == dialogic.SOURCES.TELEGRAM
    assert second['source'] == dialogic.SOURCES.TELEGRAM

    assert first['from_user'] is True
    assert second['from_user'] is False

    assert first['data'] == {'message': str(input_message)}
    assert second['data'] == tg_response

    assert first['request_id'] == second['request_id']
    assert first['request_id'] is not None
